"""G-SHAP utilities"""

import pandas as pd

def get_data(X):
    """Get data matrix

    Parameters
    ----------
    X : numpy.array or pandas.DataFrame or pandas.Series

    Returns
    -------
    (# samples x # features) numpy.array
    """
    X = X.values if isinstance(X, (pd.Series, pd.DataFrame)) else X
    return X.reshape(1, X.shape[0]) if len(X.shape)==1 else X


"""#Intergroup differences

For examples and interpretation, see my notebook on [intergroup difference explanations](https://github.com/dsbowen/gshap/blob/master/intergroup_difference.ipynb).
"""

import pandas as pd
import numpy as np
import scipy

from random import choices


class IntergroupDifference():
    """
    This class measures the distance between distributions of predicted 
    outcomes for different groups.

    Paramters
    ---------
    group : numpy.array or pandas.Series
        (# observations,) array of boolean or binary values indicating 
        group membership.

    distance : callable or str, default='absolute_mean_distance'
        Takes two vectors of model output for the outgroup and ingroup. 
        Output vectors will usually be (# outgroup,) and (# ingroup,), or 
        (# outgroup, # classes) and (# ingroup, # classes). `distance` returns
        a scalar measure of intergroup difference, such as the absolute 
        difference between group means. If input as a string, `distance` is
        used as a key to look up built-in distance functions.

    Attributes
    ----------
    group : numpy.array
        Set from the `group` parameter. If the parameter is passed as a 
        `pandas.Series`, it is automatically converted in a `numpy.array`.

    distance : callable or str
        Set from the `distance` parameter.

    Examples
    --------
    ```python
    import gshap
    from gshap.datasets import load_recidivism
    from gshap.intergroup import IntergroupDifference

    from sklearn.svm import SVC

    X, y = load_recidivism(return_X_y=True)
    clf = SVC().fit(X,y)

    g = IntergroupDifference(group=X_test['black'], distance='relative_mean_distance')
    explainer = gshap.KernelExplainer(clf.predict, X_train, g)
    explainer.gshap_values(X_test, nsamples=32)
    ```

    Out:

    ```
    array([ 0.01335252,  0.24884556,  0.00132373, -0.0025238 , -0.00151837,
    \    0.40453822,  0.01636782,  0.07666043, -0.00056414,  0.00966583])
    ```
    """
    def __init__(self, group, distance='absolute_mean_distance'):
        group = _convert_to_np(group)
        self.group = group.astype(bool)
        self.distance = (
            distance_metrics[distance] if isinstance(distance, str) else distance
        )

    def __call__(self, output):
        """
        Compute distance measure between groups.

        Parameters
        ----------
        ouput : numpy.array or pandas.Series
            Model output, usually a (# observations,) or 
            (# observations, # classes) vector.

        Returns
        -------
        distance : scalar
            Measure of the distance between the distributions of predicted 
            outputs for outgroup and ingroup observations.
        """
        out_0 = output[np.logical_not(self.group)]
        out_1 = output[self.group]
        return self.distance(out_0, out_1)


def absolute_mean_distance(out_0, out_1):
    """
    Parameters
    ----------
    out_0 : np.array
        (# observations,) vector of model outputs for outgroup observations.

    out_1 : np.array
        (# observations,) vector of model outputs for ingroup observations.

    Returns
    -------
    distance : scalar
        out_1.mean() - out_0.mean()
    """
    out_0, out_1 = [_convert_proba(vec) for vec in (out_0, out_1)]
    return out_1.mean() - out_0.mean()

def relative_mean_distance(out_0, out_1):
    """
    Parameters
    ----------
    out_0 : np.array
        (# observations,) vector of model outputs for outgroup observations.

    out_1 : np.array
        (# observations,) vector of model outputs for ingroup observations.

    Returns
    -------
    distance : scalar
        out_1.mean() / out_0.mean() - 1
    """
    out_0, out_1 = [_convert_proba(vec) for vec in (out_0, out_1)]
    return out_1.mean() / out_0.mean() - 1

def _convert_proba(vec):
    # Convert probability output from a predict_proba method to probability 
    # of being in the positive class
    if len(vec.shape) == 1:
        return vec
    return vec[:,1]

def _convert_to_np(vec):
    # Convert vector to boolean values
    if isinstance(vec, (pd.DataFrame, pd.Series)):
        return vec.values.astype(bool)
    return vec

distance_metrics = {
    'absolute_mean_distance': absolute_mean_distance,
    'relative_mean_distance': relative_mean_distance
}


"""WeightedR2 class

An accuracy metric to assess model failure.

In progress.
"""

from gshap.utils import get_data

import numpy as np
import pandas as pd

from random import choice, shuffle


class Performance():
    def __init__(self, score, model, data, y, nsamples=1000):
        self.score = score
        self.set_sample_weight(model, data, y, nsamples)

    def set_sample_weight(self, model, data, y, nsamples=1000):
        data = get_data(data)
        y = y.values if isinstance(y, (pd.DataFrame, pd.Series)) else y

        scores = []
        for i in range(nsamples):
            Z = np.array(choices(data), k=y.shape[0])
            scores.append(1/self.score(y, model(Z)))
        self.sample_weight = np.array(scores).mean(axis=0)


"""#Hypothesis testing

For examples and interpretation, see my notebook on [hypothesis test explanations](https://github.com/dsbowen/gshap/blob/master/hypothesis.ipynb).
"""

import numpy as np

from random import choices


class HypothesisTest():
    """
    This class measures how likely a hypothesis is to be true of an output 
    vector. It uses a bootstrap analysis to compute the probability that a 
    hypothesis is true of a population from a sample output vector.

    Parameters
    ----------
    test : callable
        Takes an output vector and returns a boolean indicator that the
        hypothesis is true of the output vector. This will usually involve
        computing a sample statistic of the output vector, then returning an
        indicator that the sample statistic fell within a certain range.

    bootstrap_samples : int
        Number of bootstrap samples for hypothesis testing.

    Attributes
    ----------
    test : callable
        Set from the `test` parameter.

    bootstrap_samples : int
        Set from the `bootstrap_samples` parameter.

    Examples
    --------
    ```python
    import gshap
    from gshap.hypothesis import HypothesisTest

    from sklearn.datasets import load_diabetes
    from sklearn.linear_model import Lasso

    X, y = load_diabetes(return_X_y=True)
    reg = Lasso(alpha=.1).fit(X, y)

    test = lambda y_pred: y_pred.mean() > 155
    g = HypothesisTest(test, bootstrap_samples=100)
    explainer = gshap.KernelExplainer(reg.predict, X, g)
    # artifically select a sample which with higher-than-average y
    explainer.gshap_values(X[y > 70], nsamples=100)
    ```

    Out:

    ```
    array([-0.0069,  0.0253,  0.2572,  0.1112, -0.0108, -0.0105,  0.0317,
    \    0.0009,  0.1415,  0.0071])
    ```
    """
    def __init__(self, test, bootstrap_samples=1000):
        self.test = test
        self.bootstrap_samples = bootstrap_samples

    def __call__(self, output):
        """
        Computes the probablity of the hypothesis being true of the population
        from which the sample was drawn.
        
        Parameters
        ----------
        output : numpy.array
            (# observations, # targets) vector of model outputs.

        Returns
        -------
        probability : scalar between 0 and 1
            Probability that the hypothesis is true of the population from 
            which the sample was drawn.
        """
        test_results = [
            float(self.test(np.array(choices(output, k=output.shape[0]))))
            for i in range(self.bootstrap_samples)
        ]
        return sum(test_results) / self.bootstrap_samples


"""# Example datasets"""

import pandas as pd

import os

file_dir = os.path.dirname(os.path.abspath(__file__))

def _load_dataset(filename, target, return_X_y=False):
    """
    Load a dataset.

    Parameters
    ----------
    filename : str

    target : str
        Name of target variable

    return_X_y : bool
        Indicates to return just the X and y matrices.
    
    Returns
    -------
    bunch : Bunch
        Object containing the dataframe, X feature matrix, and y target 
        vector. Or, if `return_X_y`, return (X,y).
    """
    bunch = Bunch(filename, target)
    return (bunch.data.values, bunch.target.values) if return_X_y else bunch

def load_recidivism(return_X_y=False):
    """
    Load the COMPAS recidivism dataset. The purpose of this dataset is to 
    predict whether a criminal will recidivate within two years of release.
    
    Parameters
    ----------
    return_X_y : bool, default=False
        Indicates whether to return just the X and y matrices, as opposed 
        to the data `Bunch`.

    Returns
    -------
    bunch : Bunch
        Object containing the dataframe, X feature matrix, and y target 
        vector. Or, if `return_X_y`, return (X,y).
    """
    return load_dataset(
        'compas/two-year-recidivism.csv', 
        target='two_year_recid',
        return_X_y=return_X_y
    )

def load_gdp(return_X_y=False):
    """
    Load the GDP growth dataset (from FRED data). The purpose of this 
    dataset is to forecast GDP growth based on macroeconomic variables.
    
    Parameters
    ----------
    return_X_y : bool, default=False
        Indicates whether to return just the X and y matrices, as opposed 
        to the data `Bunch`.

    Returns
    -------
    bunch : Bunch
        Object containing the dataframe, X feature matrix, and y target 
        vector. Or, if `return_X_y`, return (X,y).
    """
    return load_dataset(
        'gdp/GDP-growth.csv',
        target='GDP_g',
        return_X_y=return_X_y
    )

class Bunch():
    """
    Dataset container.

    Parameters
    ----------
    filename : str
        Name of the file in `gshap/datasets`.

    target : str
        Name of target variable

    Attributes
    ----------
    df : pandas.DataFrame
        Dataframe containing features and the target variable.

    data : pandas.DataFrame
        Dataframe containing only the features

    target : pandas.Series
        Series of the target variable.
    """
    def __init__(self, filename, target):
        self.df = pd.read_csv(os.path.join(file_dir, filename))
        self.data = self.df.drop(target, axis=1)
        self.target = self.df[target]


"""#General classification and regression explanations

For examples and interpretation, see my notebooks on [general classification explanations](https://github.com/dsbowen/gshap/blob/master/classification.ipynb) and [general regression explanations](https://github.com/dsbowen/gshap/blob/master/regression.ipynb).
"""


class ProbabilityDistance():
    """
    This class measures how likely each predicted target value (output) was to 
    have been generated by a 'positive' distribution or density, rather than a 
    'negative' distribution or density.

    Parameters
    ----------
    positive : callable or list of callables
        Densities and distributions take the output of a model, usually a 
        (# observations,) or (# observations, # classes) vector. It returns a 
        (# observations,) vector of probabilities that the predicted target 
        value was generated by the density or distribution.
        
    negative : callable or list of callables or None, default=None
        Similarly defined. If `None`, the probability that each observation 
        comes from a negative density or distribution will be treated as the 
        complement of `positive`.

    Attributes
    ----------
    positive : callable or list of callables
        Set from the `positive` parameter.

    negative : callable or list of callables
        Set from the `negative` parameter.

    Examples
    --------
    ```python
    import gshap
    from gshap.probability_distance import ProbabilityDistance

    from sklearn.datasets import load_iris
    from sklearn.svm import SVC

    X, y = load_iris(return_X_y=True)
    clf = SVC(probability=True).fit(X,y)

    # probability that each observation is in class 1
    pos_distribution = lambda y_pred: y_pred[:,1]
    # probability that each observation is in class 0
    neg_distribution = lambda y_pred: y_pred[:,0]
    g = ProbabilityDistance(pos_distribution, neg_distribution)
    explainer = gshap.KernelExplainer(clf.predict_proba, X, g)
    explainer.gshap_values(x, nsamples=1000)
    ```

    Out:

    ```
    array([0.02175944, 0.01505252, 0.17106646, 0.13605429])
    ```
    """
    def __init__(self, positive, negative=None):
        self.positive = positive
        self.negative = negative

    def __call__(self, output):
        """
        Parameters
        ----------
        output : np.array
            Model output, usually (# observations,) or 
            (# obervations, # classes) array for regression or classification 
            problems, respectively.

        Returns
        -------
        probability : float
            Probability that every predicted target value was generated by 
            a positive density or distribution, rather than a negative density 
            or distribution.
        """
        p_pos = self._compute_probability(self.positive, output)
        if self.negative:
            p_neg = self._compute_probability(self.negative, output)
        else:
            p_neg = 1 - p_pos
        x = 1 / (1 + (p_neg/p_pos).prod())
        return x

    def _compute_probability(self, funcs, output):
        """
        Compute the probability that each value of the output was generated by 
        one of the density or distribution functions.

        Parameters
        ----------
        funcs : None, callable, or list of callables
            Density functions

        output : np.array-like
            (# observations,) model output vector
        """
        if funcs is None:
            funcs = []
        elif not isinstance(funcs, list):
            funcs = [funcs]
        return sum([func(output) for func in funcs])


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gshap",
    version="0.0.1",
    author="Dillon Bowen",
    author_email="dsbowen@wharton.upenn.edu",
    description="A technique in explainable AI for answering broader questions in machine learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://dsbowen.github.io/gshap",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy>=1.18.4',
        'pandas>=1.0.3',
    ],
    python_requires='>=3.6',
)


"""# Kernel Explainer"""

from gshap.utils import get_data

from random import choices, shuffle
import numpy as np
import pandas as pd


class KernelExplainer():
    """
    The Kernel Explainer is a model-agnostic method of approximating G-SHAP 
    values.

    Parameters
    ----------
    model : callable
        Callable which takes a (# samples x # features) matrix and returns an 
        output which will be fed into `g`. For ordinary SHAP, the model 
        returns a (# samples, # targets) output vector.
    
    data : numpy.array or pandas.DataFrame or pandas.Series
        Background dataset from which values are randomly sampled to simulate 
        absent features.

    g : callable
        Callable which takes the `model` output and returns a scalar.

    Attributes
    ----------
    model : callable
        Set from the `model` parameter.

    data : numpy.array
        Set from the `data` parameter. If `data` is a `pandas` object, it is 
        automatically converted to a `numpy.array`.

    g : callable
        Set from the `g` parameter.

    Examples
    --------
    This example shows how to compute classical SHAP values.
    ```python
    import gshap
    
    from sklearn.datasets import load_boston
    from sklearn.linear_model import LinearRegression

    X, y = load_boston(return_X_y=True)
    reg = LinearRegression().fit(X,y)
    explainer = gshap.KernelExplainer(
    \    model=reg.predict, data=X, g=lambda x: x.mean()
    )
    explainer.gshap_values(X, nssamples=1000)
    ```

    Out:

    ```
    array([-8.52873964e-04, -4.90442234e-04,  9.42836482e-05,  3.98231297e-04,
    \    2.03149964e-03,  3.93086231e-03, -7.38176865e-06,  3.81400727e-03,
    \    5.19437337e-03, -1.34661588e-03,  7.08535145e-04,  1.50486721e-03,
    \   -8.28480438e-03])
    ```

    As expected, all SHAP values are 0 for linear regression. We can see this 
    when we compare the mean prediction for the original data `X` to the 
    shuffled background data `explainer.data`.

    ```python
    explainer.compare(X, bootstrap_samples=1000)
    ```

    Out:

    ```
    22.53280632411067, 22.52089950825812
    ```
    """
    def __init__(self, model, data, g=lambda x: x.mean()):
        self.model = model
        self.data = data
        self.g = g

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, X):
        self._data = (
            X.values if isinstance(X, (pd.DataFrame, pd.Series)) else X
        )
        self.N, self.P = self._data.shape

    @property
    def nsamples(self):
        """Default number of samples to draw to approximate G-SHAP values"""
        return 2 * self.P + 2**11

    def compare(self, X, bootstrap_samples=1000):
        """
        Compares the background data `self.data` to the comparison data `X` 
        in terms of the general function `self.g`.

        Parameters
        ----------
        X : numpy.array or pandas.Series or pandas.DataFrame
            (# samples, # features) matrix of comparison data.

        bootstrap_samples : scalar
            Number of bootstrapped samples for computing `g` of the 
            background data.

        Returns
        -------
        g_comparison : float
            *g(model(X))*, where *X* is the comparison data.

        g_background : float
            *g(model(X_b))*, where *X_b* is the shuffled background data.
        """
        X = get_data(X)
        g_data = []
        for i in range(bootstrap_samples):
            sample = np.array(choices(self.data, k=X.shape[0]))
            g_data.append(self.g(self.model(sample)))
        return self.g(self.model(X)), sum(g_data) / len(g_data)
        
    def gshap_values(self, X, **kwargs):
        """
        Compute G-SHAP values for all features.

        Parameters
        ----------
        X : numpy.array or pandas.DataFrame or pandas.Series
            A (# samples, # features) matrix.

        nsamples : scalar or 'auto', default='auto'
            Number of samples to draw when approximating G-SHAP values.

        Returns
        -------
        gshap_values : np.array
            (# features,) vector of G-SHAP values ordered by feature index.
        """
        return np.array(
            [self.gshap_value(j, X, **kwargs) for j in range(self.P)]
        )

    def gshap_value(self, j, X, **kwargs):
        """
        Compute the G-SHAP value for feature `j`.

        Parameters
        ----------
        j : scalar or column name
            The index or column name of the feature of interest.

        X : numpy.array or pandas.DataFrame or pandas.Series
            A (# samples, # features) matrix.

        nsamples : scalar or 'auto', default='auto'
            Number of samples to draw when approximating G-SHAP values.

        Returns
        -------
        gshap_value : float
            Approximated G-SHAP value for feature `j` (float).
        """
        j = list(X.columns).index(j) if isinstance(j, str) else j
        nsamples = kwargs.get('nsamples', self.nsamples)
        phi = [self._compute_phi(j, X) for m in range(nsamples)]
        return sum(phi) / len(phi)

    def _compute_phi(self, j, X):
        """Approximate G-SHAP value for feature `j` for one sample
        
        This method approximates the G-SHAP value by Monte Carlo sampling.
        1. Construct `Z` by sampling observations from the background dataset.
        2. Shuffle the order of the features.
        3. Construct `X_mj` (X minus the j'th feature) as all features from X 
        which come before j. Absent features are filled in from `Z`.
        4. Construct `X_pj` (X plus the j'th feature) by adding the original 
        j'th feature from `X` to `X_mj`.
        5. Return phi = g(model(X_pj)) - g(model(X_mj)).
        """
        X = get_data(X)
        # Ensure feature dimension of X matches that of the background data
        assert X.shape[1] == self.P

        Z = np.array(choices(self.data, k=X.shape[0]))

        order = np.array(list(range(self.P)))
        np.random.shuffle(order)
        j_idx = order[j]

        X_mj = (
            (order < j_idx).astype(int) * X 
            + (order >= j_idx).astype(int) * Z
        )
        X_pj = X_mj.copy()
        X_pj[:,j] = X[:,j]

        return self.g(self.model(X_pj)) - self.g(self.model(X_mj))